"""Represent a mood on the gateway."""
from .const import ROOT_MOODS
from .resource import ApiResource


class Mood(ApiResource):
    def __init__(self, raw, parent):
        super().__init__(raw)
        self._parent = parent

    @property
    def path(self):
        return [ROOT_MOODS, self._parent, self.id]

    def __repr__(self):
        return '<Mood {}>'.format(self.name)



from .const import (
    ROOT_GROUPS,
    ATTR_LIGHT_STATE,
    ATTR_LIGHT_DIMMER,
    ATTR_ID,
    ATTR_TRANSITION_TIME
)
from .resource import ApiResource

ROOT_DEVICES2 = "15002"  # ??
ATTR_MEMBERS = "9018"
ATTR_MOOD = "9039"


class Group(ApiResource):
    """Represent a group."""
    def __init__(self, gateway, raw):
        super().__init__(raw)
        self._gateway = gateway

    @property
    def path(self):
        return [ROOT_GROUPS, self.id]

    @property
    def state(self):
        """Boolean representing the light state of the group."""
        return self.raw.get(ATTR_LIGHT_STATE) == 1

    @property
    def dimmer(self):
        """Dimmer value of the group."""
        return self.raw.get(ATTR_LIGHT_DIMMER)

    @property
    def member_ids(self):
        """Members of this group."""
        info = self.raw.get(ATTR_MEMBERS, {})

        if not info or ROOT_DEVICES2 not in info:
            return []

        return info[ROOT_DEVICES2].get(ATTR_ID, [])

    @property
    def mood_id(self):
        """Active mood."""
        return self.raw.get(ATTR_MOOD)

    def members(self):
        """Return device objects of members of this group."""
        return [self._gateway.get_device(dev) for dev in self.member_ids]

    def mood(self):
        """"Active mood."""
        return self._gateway.get_mood(self.mood_id)

    def activate_mood(self, mood_id):
        """Activate a mood."""
        return self.set_values({
            ATTR_MOOD: mood_id
        })

    def set_state(self, state):
        """Set state of a group."""
        return self.set_values({
            ATTR_LIGHT_STATE: int(state)
        })

    def set_dimmer(self, dimmer, transition_time=None):
        """Set dimmer value of a group.

        dimmer: Integer between 0..255
        transition_time: Integer representing tenth of a second (default None)
        """
        values = {
            ATTR_LIGHT_DIMMER: dimmer,
        }
        if transition_time is not None:
            values[ATTR_TRANSITION_TIME] = transition_time
        return self.set_values(values)

    def __repr__(self):
        state = 'on' if self.state else 'off'
        return '<Group {} - {}>'.format(self.name, state)



"""Coap implementation using aiocoap."""
import asyncio
import json
import logging

from aiocoap import Message, Context
from aiocoap.error import RequestTimedOut, Error, ConstructionRenderableError
from aiocoap.numbers.codes import Code
from aiocoap.transports import tinydtls

from ..error import ClientError, ServerError, RequestTimeout
from ..command import Command

_LOGGER = logging.getLogger(__name__)


class PatchedDTLSSecurityStore:
    """Patched DTLS store in lieu of impl."""

    SECRET_PSK = None

    def _get_psk(self, host, port):
        return b"Client_identity", PatchedDTLSSecurityStore.SECRET_PSK


tinydtls.DTLSSecurityStore = PatchedDTLSSecurityStore


def _patched_datagram_received(self, data, addr):
    self._dtls_socket.handleMessage(self._connection, data, 0)


tinydtls.DTLSClientConnection.datagram_received = \
    _patched_datagram_received


@asyncio.coroutine
def api_factory(host, security_code, loop=None):
    """Generate a request method."""
    if loop is None:
        loop = asyncio.get_event_loop()

    PatchedDTLSSecurityStore.SECRET_PSK = security_code.encode('utf-8')

    _observations_err_callbacks = []
    _protocol = yield from Context.create_client_context(loop=loop)

    @asyncio.coroutine
    def _get_protocol():
        """Get the protocol for the request."""
        nonlocal _protocol
        if not _protocol:
            _protocol = yield from Context.create_client_context(loop=loop)
        return _protocol

    @asyncio.coroutine
    def _reset_protocol(exc):
        """Reset the protocol if an error occurs.
           This can be removed when chrysn/aiocoap#79 is closed."""
        # Be responsible and clean up.
        protocol = yield from _get_protocol()
        yield from protocol.shutdown()
        nonlocal _protocol
        _protocol = None
        # Let any observers know the protocol has been shutdown.
        nonlocal _observations_err_callbacks
        for ob_error in _observations_err_callbacks:
            ob_error(exc)
        _observations_err_callbacks.clear()

    @asyncio.coroutine
    def _get_response(msg):
        """Perform the request, get the response."""
        try:
            protocol = yield from _get_protocol()
            pr = protocol.request(msg)
            r = yield from pr.response
            return (pr, r)
        except ConstructionRenderableError as e:
            raise ClientError("There was an error with the request.", e)
        except RequestTimedOut as e:
            yield from _reset_protocol(e)
            raise RequestTimeout('Request timed out.', e)
        except Error as e:
            yield from _reset_protocol(e)
            raise ServerError("There was an error with the request.", e)

    @asyncio.coroutine
    def _execute(api_command):
        """Execute the command."""
        if api_command.observe:
            yield from _observe(api_command)
            return

        method = api_command.method
        path = api_command.path
        data = api_command.data
        parse_json = api_command.parse_json
        url = api_command.url(host)

        kwargs = {}

        if data is not None:
            kwargs['payload'] = json.dumps(data).encode('utf-8')
            _LOGGER.debug('Executing %s %s %s: %s', host, method, path, data)
        else:
            _LOGGER.debug('Executing %s %s %s', host, method, path)

        api_method = Code.GET
        if method == 'put':
            api_method = Code.PUT

        msg = Message(code=api_method, uri=url, **kwargs)

        _, res = yield from _get_response(msg)

        api_command.result = _process_output(res, parse_json)

        return api_command.result

    @asyncio.coroutine
    def request(*api_commands):
        """Make a request."""
        if len(api_commands) == 1:
            result = yield from _execute(api_commands[0])
            return result

        commands = (_execute(api_command) for api_command in api_commands)
        command_results = yield from asyncio.gather(*commands, loop=loop)

        return command_results

    @asyncio.coroutine
    def _observe(api_command):
        """Observe an endpoint."""
        duration = api_command.observe_duration
        url = api_command.url(host)
        err_callback = api_command.err_callback

        msg = Message(code=Code.GET, uri=url, observe=duration)

        # Note that this is necessary to start observing
        pr, r = yield from _get_response(msg)

        api_command.result = _process_output(r)

        def success_callback(res):
            api_command.result = _process_output(res)

        def error_callback(ex):
            err_callback(ex)

        ob = pr.observation
        ob.register_callback(success_callback)
        ob.register_errback(error_callback)
        nonlocal _observations_err_callbacks
        _observations_err_callbacks.append(ob.error)

    # This will cause a RequestError to be raised if credentials invalid
    yield from request(Command('get', ['status']))

    return request


def _process_output(res, parse_json=True):
    """Process output."""
    res_payload = res.payload.decode('utf-8')
    output = res_payload.strip()

    _LOGGER.debug('Status: %s, Received: %s', res.code, output)

    if not output:
        return None

    if not res.code.is_successful:
        if res.code >= 128 and res.code < 160:
            raise ClientError(output)
        elif res.code >= 160 and res.code < 192:
            raise ServerError(output)

    if not parse_json:
        return output

    return json.loads(output)



"""Errors for PyTradfri."""


class PyTradFriError(Exception):
    """Base Error"""
    pass


class RequestError(PyTradFriError):
    """An error happened sending or receiving a command."""
    pass


class RequestTimeout(RequestError):
    """Error when sending or receiving the command timed out."""
    pass


class ClientError(RequestError):
    """Error when the client caused the error.

    See section 5.9.2 of draft-ietf-core-coap-04.
    """
    pass


class ServerError(RequestError):
    """Error when the server caused the error.

    See section 5.9.3 of draft-ietf-core-coap-04.
    """
    pass



ROOT_DEVICES = "15001"
ROOT_GROUPS = "15004"
ROOT_MOODS = "15005"
ROOT_SMART_TASKS = "15010"
ROOT_START_ACTION = "15013"  # found under ATTR_START_ACTION
PATH_GATEWAY_INFO = ["15011", "15012"]

ATTR_APPLICATION_TYPE = "5750"
ATTR_DEVICE_INFO = "3"
ATTR_NAME = "9001"
ATTR_CREATED_AT = "9002"
ATTR_ID = "9003"
ATTR_REACHABLE_STATE = "9019"
ATTR_LAST_SEEN = "9020"
ATTR_LIGHT_CONTROL = "3311"  # array

ATTR_NTP = "9023"
ATTR_FIRMWARE_VERSION = "9029"
ATTR_CURRENT_TIME_UNIX = "9059"
ATTR_CURRENT_TIME_ISO8601 = "9060"
ATTR_FIRST_SETUP = "9069"  # ??? unix epoch value when gateway first setup
ATTR_GATEWAY_ID = "9081"  # ??? id of the gateway

ATTR_LIGHT_STATE = "5850"  # 0 / 1
ATTR_LIGHT_DIMMER = "5851"  # Dimmer, not following spec: 0..255
ATTR_LIGHT_COLOR = "5706"  # string representing a value in some color space
ATTR_LIGHT_COLOR_X = "5709"
ATTR_LIGHT_COLOR_Y = "5710"

ATTR_START_ACTION = "9042"  # array

ATTR_SMART_TASK_TYPE = "9040"  # 4 = transition | 1 = not home | 2 = on/off
ATTR_SMART_TASK_NOT_AT_HOME = 1
ATTR_SMART_TASK_LIGHTS_OFF = 2
ATTR_SMART_TASK_WAKE_UP = 4

ATTR_SMART_TASK_TRIGGER_TIME_INTERVAL = "9044"
ATTR_SMART_TASK_TRIGGER_TIME_START_HOUR = "9046"
ATTR_SMART_TASK_TRIGGER_TIME_START_MIN = "9047"

ATTR_TRANSITION_TIME = "5712"
ATTR_REPEAT_DAYS = "9041"



"""Provide a CLI for Tradfri."""
import logging
from pprint import pprint
import sys

from .const import *  # noqa
from .coap_cli import api_factory
from .gateway import Gateway
from .command import Command


if __name__ == '__main__':
    if len(sys.argv) != 3:
        print('Call with {} <host> <key>'.format(sys.argv[0]))
        sys.exit(1)

    logging.basicConfig(level=logging.DEBUG)

    api = api_factory(sys.argv[1], sys.argv[2])
    gateway = Gateway()
    devices = api(gateway.get_devices())
    lights = [dev for dev in devices if dev.has_light_control]
    light = lights[0]
    groups = api(gateway.get_groups())
    group = groups[0]
    moods = api(gateway.get_moods())
    mood = moods[0]
    tasks = api(gateway.get_smart_tasks())

    def dump_all():
        endpoints = api(gateway.get_endpoints())

        for endpoint in endpoints:
            parts = endpoint[1:].split('/')

            if not all(part.isdigit() for part in parts):
                continue

            pprint(api(Command('get', parts)))
            print()
            print()

    def dump_devices():
        pprint([d.raw for d in api(gateway.get_devices())])

    print()
    print("Example commands:")
    print("> devices")
    print("> light.light_control.lights")
    print("> light.light_control.set_dimmer(10)")
    print("> light.light_control.set_dimmer(254)")
    print("> light.light_control.set_xy_color(254)")
    print("> lights[1].light_control.set_dimmer(20)")
    print("> tasks[0].repeat_days_list")
    print("> groups")
    print("> moods")
    print("> tasks")
    print("> dump_devices()")
    print("> dump_all()")
